@@ -1,5 +1,6 @@
 from argparse import ArgumentParser, Namespace, SUPPRESS
 import attr
+import hashlib
 from pathlib import Path
 import sys
 import typing
@@ -13,39 +14,62 @@ from kgtk.value.kgtkvalue import KgtkValue
 
 @attr.s(slots=True, frozen=True)
 class KgtkIdBuilderOptions(KgtkFormat):
+    DEFAULT_OVERWRITE: bool = False
+    DEFAULT_VERIFY_ID_UNIQUE: bool = False
+
     # TODO: use an enum
     CONCAT_NLN_STYLE: str = "node1-label-node2" # node1-label-node2
     CONCAT_NL_NUM_STYLE: str = "node1-label-num" # node1-label-#
     CONCAT_NLN_NUM_STYLE: str = "node1-label-node2-num" # node1-label-node2-#
     CONCAT_WITH_OLD_ID_STYLE: str = "node1-label-node2-id" # Tag on any existing ID value
+    EMPTY_STYLE: str = "empty" # DO not generate an ID (useful for debugging and other special circumstances)
     PREFIXED_STYLE: str = "prefix###" # XXX###
+    WIKIDATA_STYLE: str = "wikidata" # node1-label-node2/hash
+    WIKIDATA_WITH_CLAIM_ID_STYLE: str = "wikidata-with-claim-id" # node1-label-node2/hash-claimidhash
+
     STYLES: typing.List[str] = [
         CONCAT_NLN_STYLE,
         CONCAT_NL_NUM_STYLE,
         CONCAT_NLN_NUM_STYLE,
         CONCAT_WITH_OLD_ID_STYLE,
+        EMPTY_STYLE,
         PREFIXED_STYLE,
+        WIKIDATA_STYLE,
+        WIKIDATA_WITH_CLAIM_ID_STYLE,
     ]
     DEFAULT_STYLE: str = PREFIXED_STYLE
     DEFAULT_CONCAT_NUM_WIDTH: int = 4
+    DEFAULT_VALUE_HASH_WIDTH: int = 6
+    DEFAULT_CLAIM_ID_HASH_WIDTH: int = 8
+    DEFAULT_CLAIM_ID_COLUMN_NAME: str = "claim_id"
 
     # Defaults for prefixed style IDs.
     DEFAULT_PREFIX: str = "E"
     DEFAULT_INITIAL_ID: int = 1
     DEFAULT_PREFIX_NUM_WIDTH: int = 1
 
+    DEFAULT_ID_SEPARATOR: str = "-"
+
     old_id_column_name: typing.Optional[str] = attr.ib(validator=attr.validators.optional(attr.validators.instance_of(str)), default=None)
     new_id_column_name: typing.Optional[str] = attr.ib(validator=attr.validators.optional(attr.validators.instance_of(str)), default=None)
-    overwrite_id: bool = attr.ib(validator=attr.validators.instance_of(bool), default=False)
-    verify_id_unique: bool = attr.ib(validator=attr.validators.instance_of(bool), default=False)
+    overwrite_id: bool = attr.ib(validator=attr.validators.instance_of(bool), default=DEFAULT_OVERWRITE)
+    verify_id_unique: bool = attr.ib(validator=attr.validators.instance_of(bool), default=DEFAULT_VERIFY_ID_UNIQUE)
     id_style: str = attr.ib(validator=attr.validators.instance_of(str), default=DEFAULT_STYLE)
     id_prefix: str = attr.ib(validator=attr.validators.instance_of(str), default=DEFAULT_PREFIX)
     initial_id: int = attr.ib(validator=attr.validators.instance_of(int), default=DEFAULT_INITIAL_ID)
     id_prefix_num_width: int = attr.ib(validator=attr.validators.instance_of(int), default=DEFAULT_PREFIX_NUM_WIDTH)
     id_concat_num_width: int = attr.ib(validator=attr.validators.instance_of(int), default=DEFAULT_CONCAT_NUM_WIDTH)
+    value_hash_width: int = attr.ib(validator=attr.validators.instance_of(int), default=DEFAULT_VALUE_HASH_WIDTH)
+    claim_id_hash_width: int = attr.ib(validator=attr.validators.instance_of(int), default=DEFAULT_CLAIM_ID_HASH_WIDTH)
+    claim_id_column_name: str = attr.ib(validator=attr.validators.instance_of(str), default=DEFAULT_CLAIM_ID_COLUMN_NAME)
+    id_separator: str = attr.ib(validator=attr.validators.instance_of(str), default=DEFAULT_ID_SEPARATOR)
 
     @classmethod
-    def add_arguments(cls, parser: ArgumentParser, expert: bool = False, overwrite: bool = False):
+    def add_arguments(cls,
+                      parser: ArgumentParser,
+                      expert: bool = False,
+                      overwrite: typing.Optional[bool] = None,
+                      default_style: str = DEFAULT_STYLE):
 
         # This helper function makes it easy to suppress options from
         # The help message.  The options are still there, and initialize
@@ -56,6 +80,8 @@ class KgtkIdBuilderOptions(KgtkFormat):
             else:
                 return SUPPRESS
 
+        if overwrite is None:
+            overwrite = cls.DEFAULT_OVERWRITE
 
         # This one is likely to cause conflicts in the future.
         #
@@ -80,9 +106,9 @@ class KgtkIdBuilderOptions(KgtkFormat):
                                   help="When true, verify ID uniqueness using an in-memory set of IDs. " +
                                   "When --verify-id-unique is omitted, it defaults to %(default)s. " +
                                   "When --verify-id-unique is supplied without an argument, it is %(const)s. ",
-                                  type=optional_bool, nargs='?', const=True, default=False)
+                                  type=optional_bool, nargs='?', const=True, default=cls.DEFAULT_VERIFY_ID_UNIQUE)
 
-        parser.add_argument(      "--id-style", dest="id_style", default=cls.DEFAULT_STYLE, choices=cls.STYLES,
+        parser.add_argument(      "--id-style", dest="id_style", default=default_style, choices=cls.STYLES,
                                   help=h("The ID generation style. (default=%(default)s)."))
 
         parser.add_argument(      "--id-prefix", dest="id_prefix", default=cls.DEFAULT_PREFIX,
@@ -101,6 +127,19 @@ class KgtkIdBuilderOptions(KgtkFormat):
                                   metavar="INTEGER",
                                   help=h("The width of the numeric value for a concatenated ID. (default=%(default)s)."))
 
+        parser.add_argument(      '--value-hash-width', action="store", type=int, dest="value_hash_width", default=cls.DEFAULT_VALUE_HASH_WIDTH,
+                                  help='How many characters should be used in a value hash? (default=%(default)d)')
+
+        parser.add_argument(     '--claim-id-hash-width', action="store", type=int, dest="claim_id_hash_width", default=cls.DEFAULT_CLAIM_ID_HASH_WIDTH,
+                                 help='How many characters should be used to hash the claim ID? 0 means do not hash the claim ID. (default=%(default)d)')
+
+        parser.add_argument(     '--claim-id-column-name', action="store", type=str, dest="claim_id_column_name", default=cls.DEFAULT_CLAIM_ID_COLUMN_NAME,
+                                 help='The name of the claim_id column. (default=%(default)s)')
+
+        parser.add_argument(     '--id-separator', action="store", type=str, dest="id_separator", default=cls.DEFAULT_ID_SEPARATOR,
+                                 help='The separator user between ID subfields. (default=%(default)s)')
+
+
     @classmethod
     def from_dict(cls, d: dict)->'KgtkIdBuilderOptions':
         return cls(
@@ -108,11 +147,15 @@ class KgtkIdBuilderOptions(KgtkFormat):
             new_id_column_name=d.get("new_id_column_name"),
             overwrite_id=d.get("overwrite_id", False),
             verify_id_unique=d.get("verify_id_unique", False),
-            id_style=d.get("id_style", cls.PREFIXED_STYLE),
+            id_style=d.get("id_style", cls.DEFAULT_STYLE),
             id_prefix=d.get("id_prefix", cls.DEFAULT_PREFIX),
             initial_id=d.get("initial_id", cls.DEFAULT_INITIAL_ID),
-            id_prefix_num_width=d.get("id_prefix_num_width", cls.DEFAULT_INITIAL_ID),
-            id_concat_num_width=d.get("id_concat_num_width", cls.DEFAULT_INITIAL_ID),
+            id_prefix_num_width=d.get("id_prefix_num_width", cls.DEFAULT_PREFIX_NUM_WIDTH),
+            id_concat_num_width=d.get("id_concat_num_width", cls.DEFAULT_CONCAT_NUM_WIDTH),
+            value_hash_width=d.get("value_hash_width", cls.DEFAULT_VALUE_HASH_WIDTH),
+            claim_id_hash_width=d.get("claim_id_hash_width", cls.DEFAULT_CLAIM_ID_HASH_WIDTH),
+            claim_id_column_name=d.get("claim_id_column_name", cls.DEFAULT_CLAIM_ID_COLUMN_NAME),
+            id_separator=d.get("id_separator", cls.DEFAULT_ID_SEPARATOR),
         )
 
     # Build the value parsing option structure.
@@ -128,10 +171,14 @@ class KgtkIdBuilderOptions(KgtkFormat):
         print("--overwrite-id=%s" % str(self.overwrite_id), file=out, flush=True)
         print("--verify_id_unique=%s" % str(self.verify_id_unique), file=out, flush=True)
         print("--id-style=%s" % str(self.id_style), file=out, flush=True)
-        print("--id-prefix=%s" % str(self.id_prefix), file=out, flush=True)
+        print("--id-prefix=%s" % repr(self.id_prefix), file=out, flush=True)
         print("--initial-id=%s" % str(self.initial_id), file=out, flush=True)
         print("--id-prefix-num-width=%s" % str(self.id_prefix_num_width), file=out, flush=True)
         print("--id-concat-num-width=%s" % str(self.id_concat_num_width), file=out, flush=True)
+        print("--value-hash-width=%s" % str(self.value_hash_width), file=out, flush=True)
+        print("--claim-id-hash-width=%s" % str(self.claim_id_hash_width), file=out, flush=True)
+        print("--claim-id-column-name=%s" % repr(self.claim_id_column_name), file=out, flush=True)
+        print("--id-separator=%s" % repr(self.id_separator), file=out, flush=True)
 
 
 @attr.s(slots=True, frozen=False)
@@ -151,6 +198,8 @@ class KgtkIdBuilder(KgtkFormat):
     label_column_idx: int= attr.ib(validator=attr.validators.instance_of(int))
     node2_column_idx: int= attr.ib(validator=attr.validators.instance_of(int))
 
+    claim_id_column_idx: int= attr.ib(validator=attr.validators.instance_of(int))
+
     current_id: int = attr.ib(validator=attr.validators.instance_of(int))
 
     id_set: typing.Set[str] =attr.ib(validator=attr.validators.instance_of(set), factory=set)
@@ -193,7 +242,7 @@ class KgtkIdBuilder(KgtkFormat):
         node1_column_idx, label_column_idx, node2_column_idx, id_column_idx = \
             KgtkReader.get_special_columns(column_name_map, "", "idbuilder")
 
-        return cls.new1(column_names, column_name_map, node1_column_idx, label_column_idx, node2_column_idx, id_column_idx, options)
+        return cls.new1(column_names.copy(), column_name_map, node1_column_idx, label_column_idx, node2_column_idx, id_column_idx, options)
 
     @classmethod
     def new1(cls,
@@ -212,7 +261,7 @@ class KgtkIdBuilder(KgtkFormat):
         if options.old_id_column_name is not None:
             # The old ID column was explicitly named.
             old_id_column_name = options.old_id_column_name
-            if new_id_column_name in column_name_map:
+            if old_id_column_name in column_name_map:
                 # The new ID column already exists.
                 old_id_column_idx = column_name_map[old_id_column_name]
             else:
@@ -258,6 +307,8 @@ class KgtkIdBuilder(KgtkFormat):
                 column_names.append(KgtkFormat.ID)
                 add_new_id_column = True
 
+        claim_id_column_idx: int = column_name_map.get(options.claim_id_column_name, -1)
+
         # Any prerequisites?
         if options.id_style == KgtkIdBuilderOptions.CONCAT_NLN_STYLE:
             if node1_column_idx < 0:
@@ -266,11 +317,13 @@ class KgtkIdBuilder(KgtkFormat):
                 raise ValueError("No label column index")
             if node2_column_idx < 0:
                 raise ValueError("No node2 column index")
+
         elif options.id_style == KgtkIdBuilderOptions.CONCAT_NL_NUM_STYLE:
             if node1_column_idx < 0:
                 raise ValueError("No node1 column index")
             if label_column_idx < 0:
                 raise ValueError("No label column index")
+
         elif options.id_style == KgtkIdBuilderOptions.CONCAT_NLN_NUM_STYLE:
             if node1_column_idx < 0:
                 raise ValueError("No node1 column index")
@@ -278,6 +331,7 @@ class KgtkIdBuilder(KgtkFormat):
                 raise ValueError("No label column index")
             if node2_column_idx < 0:
                 raise ValueError("No node2 column index")
+
         elif options.id_style == KgtkIdBuilderOptions.CONCAT_WITH_OLD_ID_STYLE:
             if node1_column_idx < 0:
                 raise ValueError("No node1 column index")
@@ -288,6 +342,23 @@ class KgtkIdBuilder(KgtkFormat):
             if old_id_column_idx < 0:
                 raise ValueError("No old ID column index")
 
+        elif options.id_style == KgtkIdBuilderOptions.WIKIDATA_STYLE:
+            if node1_column_idx < 0:
+                raise ValueError("No node1 column index")
+            if label_column_idx < 0:
+                raise ValueError("No label column index")
+            if node2_column_idx < 0:
+                raise ValueError("No node2 column index")
+        
+        elif options.id_style == KgtkIdBuilderOptions.WIKIDATA_WITH_CLAIM_ID_STYLE:
+            if node1_column_idx < 0:
+                raise ValueError("No node1 column index")
+            if label_column_idx < 0:
+                raise ValueError("No label column index")
+            if node2_column_idx < 0:
+                raise ValueError("No node2 column index")
+            if claim_id_column_idx < 0:
+                raise ValueError("No claim_id column index")
         
         return cls(options=options,
                    column_names=column_names,
@@ -299,6 +370,7 @@ class KgtkIdBuilder(KgtkFormat):
                    node1_column_idx=node1_column_idx,
                    label_column_idx=label_column_idx,
                    node2_column_idx=node2_column_idx,
+                   claim_id_column_idx=claim_id_column_idx,
                    current_id=options.initial_id
         )
 
@@ -329,12 +401,15 @@ class KgtkIdBuilder(KgtkFormat):
                 self.id_set.add(id_value)
         
 
-    def build(self, row: typing.List[str], line_number: int)->typing.List[str]:
+    def build(self,
+              row: typing.List[str],
+              line_number: int,
+              already_added: typing.Optional[bool]=False)->typing.List[str]:
         """
         Build a new ID value if needed.
         """
-        row = row.copy() # as a precaution
-        if self.add_new_id_column:
+        row = list(row).copy() # as a precaution
+        if self.add_new_id_column and not already_added:
             row.append("")
         elif self.old_id_column_idx >= 0:
             if row[self.old_id_column_idx] != "" and not self.options.overwrite_id:
@@ -359,6 +434,12 @@ class KgtkIdBuilder(KgtkFormat):
             new_id = self.build_concat_with_old_id(row)
         elif self.options.id_style == KgtkIdBuilderOptions.PREFIXED_STYLE:
             new_id = self.build_prefixed(row)
+        elif self.options.id_style == KgtkIdBuilderOptions.EMPTY_STYLE:
+            new_id = "" # Useful for debugging and other unusual circumstances.
+        elif self.options.id_style == KgtkIdBuilderOptions.WIKIDATA_STYLE:
+            new_id = self.build_wikidata_id(row)
+        elif self.options.id_style == KgtkIdBuilderOptions.WIKIDATA_WITH_CLAIM_ID_STYLE:
+            new_id = self.build_wikidata_id_with_claim_id(row)
         else:
             raise ValueError("Unknown ID style '%s'." % self.options.id_style)
 
@@ -368,32 +449,32 @@ class KgtkIdBuilder(KgtkFormat):
         return row
 
     def build_concat_nln(self, row: typing.List[str])->str:
-        return row[self.node1_column_idx] + "-" + row[self.label_column_idx] + "-" + row[self.node2_column_idx]
+        return row[self.node1_column_idx] + self.options.id_separator + row[self.label_column_idx] + self.options.id_separator + row[self.node2_column_idx]
 
     def build_concat_nl_num(self, row: typing.List[str])->str:
         # Returns node1-label-### where ### starts at 0.
         # Assumes that records are compact on (node2) or (node2, id), as needed.
-        key: str = row[self.node1_column_idx] + "-" + row[self.label_column_idx]
+        key: str = row[self.node1_column_idx] + self.options.id_separator + row[self.label_column_idx]
         if key in self.nl_keys:
             self.nl_keys[key] += 1
         else:
             self.nl_keys[key] = 0
-        return key + "-" + str(self.nl_keys[key]).zfill(self.options.id_concat_num_width)
+        return key + self.options.id_separator + str(self.nl_keys[key]).zfill(self.options.id_concat_num_width)
 
     def build_concat_nln_num(self, row: typing.List[str])->str:
         # Returns node1-label-node2-### where ### starts at 0.
         # Assumes that records are compact on (node2) or (node2, id), as needed.
-        key: str = row[self.node1_column_idx] + "-" + row[self.label_column_idx] + "-" + row[self.node2_column_idx]
+        key: str = row[self.node1_column_idx] + self.options.id_separator + row[self.label_column_idx] + self.options.id_separator + row[self.node2_column_idx]
         if key in self.nl_keys:
             self.nl_keys[key] += 1
         else:
             self.nl_keys[key] = 0
-        return key + "-" + str(self.nl_keys[key]).zfill(self.options.id_concat_num_width)
+        return key + self.options.id_separator + str(self.nl_keys[key]).zfill(self.options.id_concat_num_width)
 
     def build_concat_with_old_id(self, row: typing.List[str])->str:
-        key: str = row[self.node1_column_idx] + "-" + row[self.label_column_idx] + "-" + row[self.node2_column_idx]
+        key: str = row[self.node1_column_idx] + self.options.id_separator + row[self.label_column_idx] + self.options.id_separator + row[self.node2_column_idx]
         if row[self.old_id_column_idx] != "":
-            key += "-" + row[self.old_id_column_idx]
+            key += self.options.id_separator + row[self.old_id_column_idx]
         return key
 
     def build_prefixed(self, row: typing.List[str])->str:
@@ -401,6 +482,26 @@ class KgtkIdBuilder(KgtkFormat):
         self.current_id += 1
         return new_id
 
+    def build_wikidata_id(self, row: typing.List[str])->str:
+        node2_value: str = row[self.node2_column_idx]
+        if self.options.value_hash_width > 0 and node2_value.startswith(('L', 'P', 'Q')):
+            return self.build_concat_nln(row)
+        else:
+            return row[self.node1_column_idx] + self.options.id_separator + row[self.label_column_idx] + self.options.id_separator + \
+                hashlib.sha256(node2_value.encode('utf-8')).hexdigest()[:self.options.value_hash_width]
+
+    def build_wikidata_id_with_claim_id(self, row: typing.List[str])->str:
+        claim_id: str = row[self.claim_id_column_idx]
+        if len(claim_id) == 0:
+            # TODO: Perhaps raise an exception?
+            return self.build_wikidata_id(row)
+
+        elif self.options.claim_id_hash_width == 0:
+            return self.build_wikidata_id(row) + self.options.id_separator + claim_id.lower()
+
+        else:
+            return self.build_wikidata_id(row) + self.options.id_separator + hashlib.sha256(claim_id.lower().encode('utf-8')).hexdigest()[:self.options.claim_id_hash_width]
+
     def process(self, kr: KgtkReader, kw: KgtkWriter):
         line_number: int = 0
         row: typing.List[str]
@@ -470,5 +571,8 @@ def main():
     # Process the input file, building IDs.
     idb.process(kr, ew)
 
+    ew.close()
+    kr.close()
+
 if __name__ == "__main__":
     main()
